
/* Parseur pour le compilateur Pascal */

%{

  open Ast

  (* Construction d'expressions et d'instructions localises *)

  let loc () = symbol_start_pos (), symbol_end_pos ()
  let dummy_loc = Lexing.dummy_pos, Lexing.dummy_pos

  let mk_expr e = { p_expr_node = e; p_expr_loc = loc () }
    
  let mk_st s = { p_st_node = s; p_st_loc = loc () }

  let mk_st_i i s = 
    { p_st_node = s; p_st_loc = rhs_start_pos i, rhs_end_pos i }

  let mk_pat p = { p_pattern_node = p; p_pattern_loc = loc () }

%}

%token <string> IDENT
%token <int> CONST_INTEGER
%token <float> CONST_REAL

/* Mots cls */

%token PROGRAM VAR BOOLEAN INTEGER REAL ARRAY OF BEGIN END 
%token TRUE FALSE FUNCTION PROCEDURE AND OR NOT
%token IF THEN ELSE TYPE CASE WHILE DO DIV MOD FOR TO

/* Symboles */

%token LPAR RPAR LSQUARE RSQUARE UNDERSCORE
%token SEMICOLON COLON COMMA DOT DOTDOT COLONEQUAL BAR EQUAL 
%token EOF

/* Oprateurs */

%token <Ast.binop> COMP 
%token PLUS MINUS 
%token STAR SLASH

/*s Prcdences */

%nonassoc THEN
%nonassoc ELSE

%left COMP EQUAL             /* < <= > >= <> = */
%left PLUS MINUS OR          /* + - or */
%left STAR SLASH DIV MOD AND /* * / div mod and */
%nonassoc uminus uplus          /* + - */
%nonassoc NOT                /* not */

/*s Point d'entre */

%start file
%type <Ast.p_file> file

%%

file:
| PROGRAM ident SEMICOLON opt_types opt_vars list0_definitions bloc DOT EOF 
   { { p_name = $2;
       p_decl_types = $4;
       p_decl_vars = $5;
       p_definitions = $6;
       p_main = mk_st_i 7 (PSbloc $7) } } 
;

opt_types:
  /* epsilon */      { [] }
| TYPE list1_typedef { List.rev $2 }
;

list1_typedef:
  typedef               { [$1] }
| list1_typedef typedef { $2 :: $1 }
;

typedef:
  ident EQUAL list1_constructor_sep_bar SEMICOLON { $1, List.rev $3 }
;

list1_constructor_sep_bar:
  constructor                               { [$1] }
| list1_constructor_sep_bar BAR constructor { $3 :: $1 }
;

constructor:
  ident                             { $1, [] }
| ident OF list1_type_expr_sep_star { $1, List.rev $3 }
;

list1_type_expr_sep_star:
  type_expr                               { [$1] }
| list1_type_expr_sep_star STAR type_expr { $3 :: $1 }
;

opt_vars:
  /* epsilon */      { [] }
| VAR list1_var_decl { $2 }
;

list1_var_decl: 
| var_decl                { $1 }
| list1_var_decl var_decl { $1 @ $2 }
;

var_decl:
  list1_ident_sep_comma COLON type_expr SEMICOLON 
    { List.rev_map (fun v -> v, $3) $1 }
;

list1_definitions:
  definition                   { [$1] }
| list1_definitions definition { $2 :: $1 }
;

definition:
  FUNCTION ident parameters COLON type_expr SEMICOLON opt_vars bloc SEMICOLON 
    { ($2, $3, PDfunction $5, $7, mk_st_i 8 (PSbloc $8)) }
| PROCEDURE ident parameters SEMICOLON opt_vars bloc SEMICOLON 
    { ($2, $3, PDprocedure, $5, mk_st_i 6 (PSbloc $6)) }
;

parameters:
  /* epsilon */                           { [] }
| LPAR list1_parameter_sep_semicolon RPAR { $2 }
;

list1_parameter_sep_semicolon:
  parameter                                         { $1 }
| list1_parameter_sep_semicolon SEMICOLON parameter { $1 @ $3 }
;

parameter:
  list1_ident_sep_comma COLON type_expr 
    { List.rev_map (fun s -> Value, s, $3) $1 }
| VAR list1_ident_sep_comma COLON type_expr
    { List.rev_map (fun s -> Reference, s, $4) $2 }
;

list0_definitions:
  /* epsilon */     { [] }
| list1_definitions { List.rev $1 }
;

constant:
  TRUE          { Cboolean true }
| FALSE         { Cboolean false }
| CONST_INTEGER { Cinteger $1 }
| CONST_REAL    { Creal $1 }
;

expression:
  constant { mk_expr (PEconst $1) }
| left_value { mk_expr (PElvalue $1) }
| expression EQUAL expression { mk_expr (PEbinop (Beq, $1, $3)) }
| expression COMP expression { mk_expr (PEbinop ($2, $1, $3)) }
| expression PLUS expression { mk_expr (PEbinop (Badd, $1, $3)) }
| expression MINUS expression { mk_expr (PEbinop (Bsub, $1, $3)) }
| expression STAR expression { mk_expr (PEbinop (Bmul, $1, $3)) }
| expression SLASH expression { mk_expr (PEbinop (Bslash, $1, $3)) }
| expression DIV expression { mk_expr (PEbinop (Bdiv, $1, $3)) }
| expression MOD expression { mk_expr (PEbinop (Bmod, $1, $3)) }
| expression AND expression { mk_expr (PEbinop (Band, $1, $3)) }
| expression OR expression { mk_expr (PEbinop (Bor, $1, $3)) }
| MINUS expression %prec uminus { mk_expr (PEunop (Uminus, $2)) }
| PLUS expression %prec uplus { mk_expr (PEunop (Uplus, $2)) }
| NOT expression %prec uplus { mk_expr (PEunop (Unot, $2)) }
| LPAR expression RPAR { $2 }
| ident LPAR arguments RPAR 
   { mk_expr (PEcall ($1, $3)) }
;

arguments:
  list1_expression_sep_comma { List.rev $1 }
;

list1_expression_sep_comma:
  expression                                  { [$1] }
| list1_expression_sep_comma COMMA expression { $3 :: $1 }
;

left_value:
  ident                                 { PLident $1 }
| left_value LSQUARE expression RSQUARE { PLarrget ($1, $3) }
;

statement:
| left_value COLONEQUAL expression
    { mk_st (PSassign ($1, $3)) }
| ident LPAR arguments RPAR
    { mk_st (PSprocedure_call ($1, $3)) }
| ident 
    { mk_st (PSprocedure_call ($1, [])) }
| IF expression THEN statement
    { mk_st (PSif ($2, $4, mk_st PSskip)) }
| IF expression THEN statement ELSE statement 
    { mk_st (PSif ($2, $4, $6)) }
| bloc
    { mk_st (PSbloc $1) }
| WHILE expression DO statement
    { mk_st (PSwhile ($2, $4)) }
| FOR ident COLONEQUAL expression TO expression DO statement
    { mk_st (PSfor ($2, $4, $6, $8)) }
| CASE expression OF branches END
    { mk_st (PSmatch ($2, $4)) } 
;

branches:
  list1_branch_sep_bar     { List.rev $1 }
;

list1_branch_sep_bar:
  branch                          { [$1] }
| list1_branch_sep_bar BAR branch { $3 :: $1 }
;

branch:
  pattern COLON bloc_contents { $1, $3 }
;

pattern:
  UNDERSCORE                              { mk_pat PPall }
| constant                                { mk_pat (PPconst $1) }
| ident                                   { mk_pat (PPid $1) }
| ident LPAR list1_pattern_sep_comma RPAR 
    { mk_pat (PPconstructor ($1, List.rev $3)) } 
;

list1_pattern_sep_comma:
  pattern                               { [$1] }
| list1_pattern_sep_comma COMMA pattern { $3 :: $1 }
;

bloc:
| BEGIN bloc_contents END { $2 }
;

bloc_contents:
  /* epsilon */ { [] }
| SEMICOLON     { [] }
| list1_statement_sep_semicolon opt_semicolon { List.rev $1 }
;

opt_semicolon:
  /* epsilon*/ { () }
| SEMICOLON    { () }
;

list1_statement_sep_semicolon:
  statement                                         { [$1] }
| list1_statement_sep_semicolon SEMICOLON statement { $3 :: $1 }
;

type_expr:
  BOOLEAN { PTboolean }
| INTEGER { PTinteger }
| REAL    { PTreal }
| ARRAY LSQUARE CONST_INTEGER DOTDOT CONST_INTEGER RSQUARE OF type_expr
          { PTarray (loc(), $3, $5, $8) }
| ident   { PTid $1 }
;

list1_ident_sep_comma:
  ident                             { [$1] }
| list1_ident_sep_comma COMMA ident { $3 :: $1 }
;

ident:
| IDENT { { id_name = $1; id_loc = loc () } }
;
